/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Dept. of Computer Science & Engineering, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#include "cudautils.h"
#include "ptx.h"
#include "context.h"
#include "stream.h"
#include "set.h"

int available_cuda_gpus;     /* Number of CUDA-capable GPUs with SM >= 3.5 */
cuda_dev_t *cuda_gpus;       /* Global GPU bookkeeping array */

void __dbg(const char * format, ... )
{
#if defined(DEBUG)
	va_list args;
	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
#endif
}

void __check_cuda_errors(char *func, CUresult err, const char *file, const int line, 
						int destroy_context, cuda_dev_t *cuda_gpu)
{
	const char *error_desc;
	
	if (err != CUDA_SUCCESS) 
	{
		cuGetErrorString(err, &error_desc);
		fprintf(stderr,
				"[%s] %s: %s:%i: CUDA error: %s; exiting.\n",
				modulename, func, file, line, error_desc );
		if (destroy_context && cuda_gpu)
			cuda_ctx_test_destroy(cuda_gpu);
		exit(1);
	}
}

/**
 * Creates a new copy of src
 */
static
void *memcpy_new(void *src, size_t bytes)
{
	void *copy = (void*) malloc(bytes);
	memcpy(copy, src, bytes);
	return copy;
}

/**
 * Calculates the number of CUDA cores for given SM 
 */
static
int get_sm_num_cores(int major, int minor) {
	typedef struct {
		int sm;
		int cores;
	} sm_to_cores;

	/* 
	 * https://en.wikipedia.org/wiki/CUDA
	 * Number of ALU lanes for integer and single-precision 
	 * floating-point arithmetic operations.
	 */
	sm_to_cores num_gpu_arch_cores_per_sm[] = {
		{0x30, 192}, {0x32, 192}, {0x35, 192}, {0x37, 192}, 
		{0x50, 128}, {0x52, 128}, {0x53, 128}, {0x60,  64}, 
		{0x61, 128}, {0x62, 128}, {0x70,  64}, {0x72,  64}, 
		{0x75,  64}, {0x80,  64}, {-1, -1}
	};

	int index = 0;

	while (num_gpu_arch_cores_per_sm[index].sm != -1) {
		if (num_gpu_arch_cores_per_sm[index].sm == ((major << 4) + minor))
			return num_gpu_arch_cores_per_sm[index].cores;

		index++;
	}
		
	return num_gpu_arch_cores_per_sm[index-1].cores;
}	

/** 
 * Translates a host address to the corresponding CUDA mediary one
 */
static
CUdeviceptr get_host2med_addr(void *hostaddr, cuda_dev_t *cuda_gpu)
{
	setelem(hostmed_addrmap) e;

	if ((e = set_get(cuda_gpu->addr_map, hostaddr)) == NULL)
		return 0;

	return e->value;
}

/**
 * Translates a CUDA mediary address to the corresponding host one 
 */ 
static
void *get_med2host_addr(CUdeviceptr medaddr, cuda_dev_t *cuda_gpu)
{
	setelem(hostmed_addrmap) e;

	for (e = cuda_gpu->addr_map->first; e; e = e->next)
		if (e->value == medaddr)
			return e->key;

	return NULL;
}

/**
 * Copies host memory to newly allocated CUDA global memory
 */
static
void memcpy_new_host2dev(void *hostaddr, CUdeviceptr **devaddr, size_t size, 
	                     cuda_dev_t *cuda_gpu)
{
	LOGFUNC("memcpy_new_host2dev");
	CUstream *stream = cuda_stream_get(cuda_gpu);
	cuda_do(cuMemAlloc(*devaddr, size));
	cuda_do(cuMemcpyHtoDAsync(**devaddr, hostaddr, size, *stream));
}

/**
 * Calculates the two optimal dimensions of the execution
 * grid and block, given a team and a thread size.
 */
static
void set_2D_from_1D_kerneldims(int size, cuda_dim_t *dims)
{
	int min = INT_MAX, i;

	for (i = 1; i * i <= size; i++) 
	{
		if (size % i == 0)
		{
			if (size/i - i < min)
			{
				dims->X = i;
				dims->Y = size/i;
				min = (size/i) - i;
			}
		}
	}
		
	/* Currently we are 2D ... */
	dims->Z = 1;
} 

static
void set_dims(cuda_dim_t *block_dims, int num_1d, unsigned long long dims)
{
	unsigned long x, y, z;

	_ull_decode3(dims, &x, &y, &z);
	block_dims->X = x ? x : num_1d;
	block_dims->Y = y ? y : 1;
	block_dims->Z = z ? z : 1;
}
/** 
 * Calculates number of available GPUs (SM >= 3.5)
 */
int cuda_get_num_gpus(void)
{
	LOGFUNC("cuda_get_num_gpus");
	CUdevice device;
	CUresult err;
	int i = 0;
	int device_count = 0;
	int sm_major = 0, sm_minor = 0;
	int cuda_gpus = 0;

	cuda_do(cuInit(0));
	cuda_do(cuDeviceGetCount(&device_count));
	
	if (device_count == 0) {
		fprintf(stderr, 
			"cuda_get_num_gpus(): error: no CUDA-capable devices found\n");
		exit(1);
	}

	for (i = 0; i < device_count; i++)
	{
		cuda_do(cuDeviceGet(&device, i));
		cuda_do(cuDeviceGetAttribute(&sm_major, 
									CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR, 
									device));

		cuda_do(cuDeviceGetAttribute(&sm_minor, 
									CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR, 
									device));
		
		/* Count only GPUs that have compute capability >= 3.5 */
		if ((sm_major << 4) + sm_minor >= 0x35);
			cuda_gpus++;
	}

	return cuda_gpus;
}

/**
 * Initializes the information structure of every CUDA-capable GPU.
 */
void cuda_init(void)
{
	LOGFUNC("cuda_init");
	int i;

	if (cuda_gpus == NULL)
	{
		if (available_cuda_gpus == 0)
			available_cuda_gpus = cuda_get_num_gpus();
		cuda_gpus = (cuda_dev_t*) malloc(available_cuda_gpus*sizeof(cuda_dev_t));
	}

	for (i = 0; i < available_cuda_gpus; i++)
	{
		/* Get device with offset `i` */
		cuda_do(cuDeviceGet(&(cuda_gpus[i].device), i));
		
		/* Get SM version major and minor */
		cuda_do(cuDeviceGetAttribute(&(cuda_gpus[i].device_sm_version_major), 
									CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR, 
									cuda_gpus[i].device));

		cuda_do(cuDeviceGetAttribute(&(cuda_gpus[i].device_sm_version_minor), 
									CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR, 
									cuda_gpus[i].device));
							
		/* Get number of multiprocessors for the selected device */
		cuda_do(cuDeviceGetAttribute(&(cuda_gpus[i].device_num_mp), 
									CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, 
									cuda_gpus[i].device));
							
		/* Get maximum number of threads per block */
		cuda_do(cuDeviceGetAttribute(&(cuda_gpus[i].device_max_thread_block_size), 
									CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK, 
									cuda_gpus[i].device));

		/* Get maximum number of threads per block */
		cuda_do(cuDeviceGetAttribute(&(cuda_gpus[i].device_max_blocks_per_grid), 
									CU_DEVICE_ATTRIBUTE_MAX_GRID_DIM_X, 
									cuda_gpus[i].device));									
												
		/* Get device global memory size */										
		cuda_do(cuDeviceTotalMem(&(cuda_gpus[i].device_global_mem_size), 
								cuda_gpus[i].device));
												
		/* Get device name */
		cuda_do(cuDeviceGetName(cuda_gpus[i].device_name, 256, 
								cuda_gpus[i].device));
		
		/* Get number of cores per multiprocessor, based on the SM */
		cuda_gpus[i].device_num_mp_cores = get_sm_num_cores(
								cuda_gpus[i].device_sm_version_major, 
								cuda_gpus[i].device_sm_version_minor);

		pthread_key_create(&(cuda_gpus[i].ctx_key), NULL);
		pthread_key_create(&(cuda_gpus[i].stream_key), NULL);

#if defined(USE_CACHING)
		pthread_key_create(&(cuda_gpus[i].cache_key), NULL);
#endif
	}
}

/**
 * Clears global GPU bookkeeping array
 */
void cuda_clear(void)
{
	int i;
	if (cuda_gpus == NULL)
		return;

	for (i = 0; i < available_cuda_gpus; i++)
	{
		pthread_key_delete(cuda_gpus[i].ctx_key);
		pthread_key_delete(cuda_gpus[i].stream_key);
#if defined(USE_CACHING)
		pthread_key_delete(cuda_gpus[i].cache_key);
#endif
	}

	free(cuda_gpus);
	cuda_gpus = NULL;
}

#define MODSUB(x,c) (((x)+CUDA_CACHE_SIZE-(c)) % CUDA_CACHE_SIZE)

/**
 * Tests if a thread-specific kernel cache is set. If not,
 * it initializes and sets a new one.
 */
void cuda_cache_test_init(cuda_dev_t *cuda_gpu)
{
#if defined(USE_CACHING)
	int i;
	cuda_kernel_cache_t *cache = 
		(cuda_kernel_cache_t*) pthread_getspecific(cuda_gpu->cache_key);

	if (cache == NULL)
	{
		cache = (cuda_kernel_cache_t*) malloc(sizeof(cuda_kernel_cache_t));
		cache->nkernels = 0;
		for (i = 0; i < CUDA_CACHE_SIZE; i++)
		{
			cache->kernels[i].name = NULL;
			cache->kernels[i].nargs = 0;
		}
		pthread_setspecific(cuda_gpu->cache_key, cache);
	}
#endif
}


/**
 * Tests if a thread-specific kernel cache is set. If so,
 * it destroys it and frees it.
 */
void cuda_cache_test_destroy(cuda_dev_t *cuda_gpu)
{
#if defined(USE_CACHING)
	cuda_kernel_cache_t *cache = 
		(cuda_kernel_cache_t*) pthread_getspecific(cuda_gpu->cache_key);

	if (cache != NULL)
		free(cache);
#endif
}

/**
 * Copies a variable to a symbol that resides in CUDA constant memory.
 */
void cuda_memcpy_const(cuda_dev_t *cuda_gpu, char *symbol, void *src, size_t bytes)
{
	LOGFUNC("cuda_memcpy_const");
	CUdeviceptr devptr;
	CUstream *stream = cuda_stream_get(cuda_gpu);
	cuda_do(cuModuleGetGlobal(&devptr, NULL, cuda_gpu->module, symbol));
	cuda_do(cuMemcpyHtoDAsync(devptr, src, bytes, *stream));
}

/** 
 * Retrieves a CUDA kernel from the kernel cache,
 * otherwise creates & caches it.
 */
int cuda_kernel_cached_create(cuda_dev_t *cuda_gpu, void *(*host_func)(void *), 
	                          char *kernel_filename, char *kernel_name)
{
	int i, nkernels, kernel_id = 0;
	bool cached = false;
	cuda_kernel_t *kernel;
	
	LOGFUNC("cuda_kernel_cached_create");

#if defined(USE_CACHING)
	cuda_kernel_cache_t *cache = 
		(cuda_kernel_cache_t*) pthread_getspecific(cuda_gpu->cache_key);
	nkernels = cache->nkernels;

	/* We suppose that the kernel is not cached, thus
	given a new id */
	kernel_id = nkernels % CUDA_CACHE_SIZE;

	if (nkernels <= CUDA_CACHE_SIZE)
	{
		for (i = 0; i < nkernels; i++)
		{
			if (cache->kernels[i].host_func == host_func)
			{
				kernel_id = i;
				cached = true;
			}
		}
	}
	else
	{
		for (i = 1; i <= CUDA_CACHE_SIZE; i++)
		{
			if (cache->kernels[MODSUB(nkernels, i)].host_func == host_func)
			{
				kernel_id = MODSUB(nkernels, i);
				cached = true;
			}
		}
	}
		
	kernel = (cuda_kernel_t*) &(cache->kernels[kernel_id]);
#else
	kernel = cuda_gpu->kernel;
#endif

	if (!cached)
	{
		kernel->nargs = 0;
		kernel->name = strdup(kernel_name);
		kernel->filename = strdup(kernel_filename);
		kernel->host_func = host_func;
		kernel->libdevpart = (char *) malloc(256); // TODO memory leak
		
#if defined(USE_CACHING)
		cache->nkernels++;
#endif

#if defined(CUDA_JIT_COMPILATION)
		ptx_compile_and_load(&(cuda_gpu->module), kernel);
#else
		cuda_do_crit(cuModuleLoad(&(cuda_gpu->module), kernel_filename), cuda_gpu);
		cuda_do_crit(cuModuleGetFunction(&(kernel->function), cuda_gpu->module, 
				kernel->name), cuda_gpu); 
#endif
	}

	return kernel_id;
}

/**
 * Prepares CUDA kernel arguments for a GPU
 */ 
void cuda_kernel_set_args(cuda_dev_t *cuda_gpu, void *dev_data, int kernel_id,
							int *num_args, void **args) 
{
	LOGFUNC("cuda_kernel_set_args");
	void *new_kernel_args[256];  /* Array used for storing kernel args temporarily */
	void *ptr_var;
	unsigned long offset_var;
	int ndeclargs, nfirstprivate, i, argoffset, sizeoffset;
	CUdeviceptr *device_struct;
	cuda_kernel_t *kernel;
	size_t totalargs;
	int off = 0;

#if defined(USE_CACHING)
	cuda_kernel_cache_t *cache = 
		(cuda_kernel_cache_t*) pthread_getspecific(cuda_gpu->cache_key);

	if (kernel_id >= cache->nkernels)
	{
		fprintf(stderr, "%s: error: invalid kernel ID", __currfunc__);
		exit(1);
	}
	kernel = (cuda_kernel_t*) &(cache->kernels[kernel_id]);
#else
	kernel = cuda_gpu->kernel;
#endif

	device_struct = (CUdeviceptr*) malloc(sizeof(CUdeviceptr));
	
	/* Handle target declare arguments */
	ndeclargs = num_args[ARGS_NUMDECL];

	for (i = 0; i < ndeclargs; i++, kernel->nargs++)
	{
		ptr_var = (void*) args[off++];
		new_kernel_args[i+1] = 
			(void*) memcpy_new(&ptr_var, sizeof(CUdeviceptr));
	}
	
	if (dev_data != NULL)
	{
		/* If num_mapped_allocs == 2, then both dev and decl data 
		exist. Size of dev data is stored in the 2nd position
		of sizes array. */
		sizeoffset = (cuda_gpu->num_mapped_allocs == 2);

		/* Allocate memory for dev data struct and copy it to device */
		memcpy_new_host2dev(dev_data, &device_struct,
		                     cuda_gpu->mapped_memory_sizes[sizeoffset], 
		                     cuda_gpu);

		/* Handle firstprivate arguments */
		nfirstprivate = num_args[ARGS_NUMFIP];
		
		for (i = ndeclargs; i < ndeclargs + nfirstprivate; i++, kernel->nargs++)
		{
			ptr_var = (void *) args[off++];
			new_kernel_args[i+1] = 
				(void*) memcpy_new(&ptr_var, sizeof(CUdeviceptr));
		}

		/* Prepare the rest of the kernel arguments */
		ptr_var = (void *) args[off++];
		for (i = 0; i < num_args[ARGS_NUMMAPPED]; i++, kernel->nargs++)
		{
			argoffset = ndeclargs + nfirstprivate + i + 1;
			if ((i % 2) == 1)
			{
				/* As far as offsets are concerned, we just need
				to pass them by value. Allocating memory on host and 
				copying the var arg is the minimum we can do here. */
				new_kernel_args[argoffset] = (void*) memcpy_new(&offset_var, 
					sizeof(unsigned long));
				ptr_var = (void*) args[off++];
			}
			else
			{
				/* Memory for the kernel arguments has already been allocated 
				in the device. ptr_var is a mediary address, however keeping
				a copy on host is necessary for offloading. */
				new_kernel_args[argoffset] = (void*) memcpy_new(&ptr_var, 
					sizeof(CUdeviceptr));
				offset_var = (unsigned long) args[off++];
			}
		}
	}
	else 
	{
		/* If there was no dev data struct, allocate 1 byte of memory
		and pass it to the kernel, as kernels do not allow NULL
		parameters. */
		cuda_do(cuMemAlloc(device_struct, 1));
	}

	new_kernel_args[0] = (void*) device_struct;
	kernel->nargs++;

	/* Copy new arguments */
	kernel->args = (void**) malloc(kernel->nargs*sizeof(void*));

	for (i = 0; i < kernel->nargs; i++)
	{
		kernel->args[i] = new_kernel_args[i];
		new_kernel_args[i] = NULL;
	}
}

/** 
 * Launches a CUDA kernel
 */
void cuda_kernel_launch(cuda_dev_t *cuda_gpu, int kernel_id, int num_teams, int num_threads, 
						unsigned long long teamdims, unsigned long long thrdims, 
						int thread_limit)
{
	LOGFUNC("cuda_kernel_launch");
	int i;
	cuda_kernel_t *kernel;
	cuda_dim_t *grid_dims, *block_dims;

#if defined(USE_CACHING)
	cuda_kernel_cache_t *cache = 
		(cuda_kernel_cache_t*) pthread_getspecific(cuda_gpu->cache_key);

	if (kernel_id >= cache->nkernels)
	{
		fprintf(stderr, "%s: error: invalid kernel ID", __currfunc__);
		exit(1);
	}
	
	kernel = (cuda_kernel_t*) &(cache->kernels[kernel_id]);
#else
	kernel = cuda_gpu->kernel;
#endif

	grid_dims = &(kernel->grid_dims);
	block_dims = &(kernel->block_dims);
	
	set_dims(grid_dims, num_teams, teamdims);
	set_dims(block_dims, num_threads, thrdims);
 
	DBG(("num_teams %d - num_threads %d - thread_limit %d\n", 
		num_teams, num_threads, thread_limit));

	DBG(("grid size: %d x %d x %d\n", 
		grid_dims->X, grid_dims->Y, grid_dims->Z));

	DBG(("block size: %d x %d x %d\n", 
		block_dims->X, block_dims->Y, block_dims->Z));
		
	/* Launch kernel with specified grid and block dimensions */
	if (cuda_gpu->device_sm_version_major >= 6)
	{
		cuda_do_crit(cuLaunchCooperativeKernel(kernel->function, 
	             grid_dims->X, grid_dims->Y, grid_dims->Z,
	             block_dims->X, block_dims->Y, block_dims->Z,
	             CUDA_SHMEM_SIZE, 0, kernel->args), cuda_gpu);
	}
	else
	{
		cuda_do_crit(cuLaunchKernel(kernel->function, 
		             grid_dims->X, grid_dims->Y, grid_dims->Z,
		             block_dims->X, block_dims->Y, block_dims->Z,
		             CUDA_SHMEM_SIZE, 0, kernel->args, 0), cuda_gpu);
	}
	
	cuCtxSynchronize();

	for (i = 0; i < kernel->nargs; i++)
		free(kernel->args[i]);
	free(kernel->args);
	kernel->nargs = 0;

	cuda_gpu->num_launched_kernels++;
}

#if defined(DEBUG)
#undef DEBUG
#endif
